#include <fcntl.h>
#include "linux/i2c-dev.h"
#include <stdio.h>
#include <sys/mman.h>
#include <errno.h>
#include <linux/types.h>
#include <sys/ioctl.h>

#define FS454_I2C_DEV	"/dev/i2c/2"
#define STV6412_I2C_DEV	"/dev/i2c/1"
#define FS454_I2C_ADDR	0x6A // SMT bus offset adress

static const unsigned char fs454_defaults[] = {
		0x64, 0x00, 0x1a, 0x00, 0xd0, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x21, 0x10, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x60, 0x72, 0x7c, 0x00, 0x04, 0x04, 0x1f, 0x03,
		0x00, 0x00, 0x40, 0x00, 0x0a, 0x00, 0x10, 0x00, 0x0d, 0x01, 0x00, 0x00, 0x00, 0x00, 0x08, 0x11,
		0x00, 0x0f, 0x05, 0xfe, 0x0b, 0x80, 0x00, 0x00, 0x82, 0x00, 0x0c, 0x02, 0x00, 0x00, 0x01, 0x00,
		0x2a, 0x09, 0x8a, 0xcb, 0x00, 0x00, 0x8d, 0x08, 0x7c, 0x3c, 0x9a, 0x2f, 0x21, 0x01, 0x3f, 0x00,
		0x3e, 0x03, 0x1a, 0x2a, 0x1c, 0x3d, 0x14, 0x9c, 0x01, 0x00, 0xfe, 0x7e, 0x60, 0x05, 0xc8, 0x00,
		0x9c, 0x07, 0x9c, 0x00, 0x00, 0x00, 0x08, 0x04, 0x00, 0x1a, 0x55, 0x01, 0x1e, 0x07, 0x7e, 0x02,
		0x54, 0xb1, 0x04, 0x14, 0x69, 0x10, 0x00, 0xa3, 0xc8, 0x22, 0x02, 0x22, 0x3e, 0x03, 0x00, 0x20,
		0x07, 0x2f, 0x07, 0x18, 0x00, 0x00, 0x18, 0x00, 0x00, 0x13, 0x13, 0x9c, 0x03, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x01, 0x0b, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xe4, 0x00,
		0x00, 0x02, 0x4d, 0x00, 0x96, 0x00, 0x1d, 0x00, 0xa0, 0x00, 0xdb, 0x00, 0x7e, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00
	};

/*
 *
 * FS454 registers
 *
 */
#define FS454_IHO		0x00	/*! Input Horizontal Offset */
#define FS454_IVO		0x02	/*! Input Vertical Offset */
#define FS454_IHW		0x04	/*! Input Horizontal Width */
#define FS454_VSC		0x06	/*! Vertical Scaling Coefficient */
#define FS454_HSC		0x08	/*! Horizontal Scaling Coefficient */
#define FS454_BYPASS		0x0A	/*! BYPASS */
#define FS454_CR		0x0C	/*! Command Register */
#define FS454_MISC		0x0E	/*! Miscellaneous Bits Register*/
#define FS454_NCON		0x10	/*! Numerator of NCO Word */
#define FS454_NCOD		0x14	/*! Denominator of NCO Word */
#define FS454_PLL_M_PUMP	0x18	/*! PLL M and Pump Control */
#define FS454_PLL_N		0x1A	/*! PLL N */
#define FS454_PLL_PDIV		0x1C	/*! PLL Post-Divider */
#define FS454_SHP		0x24	/*! Sharpness Filter */
#define FS454_FLK		0x26	/*! Filcker Filter Coefficient */
#define FS454_GPIO		0x28	/*! General Purpose I/O, Output Enab */
#define FS454_ID		0x32	/*! Part Identification Number */
#define FS454_STATUS		0x34	/*! Status Port */
#define FS454_FIFO_SP		0x36	/*! FIFO Status Port Fill/Underrun */
#define FS454_FIFO_LAT		0x38	/*! FIFO Latency */
#define FS454_CHR_FREQ		0x40	/*! Chroma Subcarrier Frequency */
#define FS454_CHR_PHASE		0x44	/*! Chroma Phase */
#define FS454_MISC_45		0x45	/*! Miscellaneous Bits Register 45 */
#define FS454_MISC_46		0x46	/*! Miscellaneous Bits Register 46 */
#define FS454_MISC_47		0x47	/*! Miscellaneous Bits Register 47 */
#define FS454_HSYNC_WID		0x48	/*! HSync Width */
#define FS454_BURST_WID		0x49	/*! Burst Width */
#define FS454_BPORCH		0x4A	/*! Back Porch Width */
#define FS454_CB_BURST		0x4B	/*! Cb Burst Amplitude */
#define FS454_CR_BURST		0x4C	/*! Cr Burst Amplitude */
#define FS454_MISC_4D		0x4D	/*! Miscellaneous Bits Register 4D */
#define FS454_BLACK_LVL		0x4E	/*! Black Level */
#define FS454_BLANK_LVL		0x50	/*! Blank Level */
#define FS454_NUM_LINES		0x57	/*! Number of Lines */
#define FS454_WHITE_LVL		0x5E	/*! White Level */
#define FS454_CB_GAIN		0x60	/*! Cb Color Saturation */
#define FS454_CR_GAIN		0x62	/*! Cr Color Saturation */
#define FS454_TINT		0x65	/*! Tint */
#define FS454_BR_WAY		0x69	/*! Width of Breezeway */
#define FS454_FR_PORCH		0x6C	/*! Front Porch */
#define FS454_NUM_PIXELS	0x71	/*! Total num. of luma/chroma Pixels */
#define FS454_1ST_LINE		0x73	/*! First Video Line */
#define FS454_MISC_74		0x74	/*! Miscellaneous Bits Register 74 */
#define FS454_SYNC_LVL		0x75	/*! Sync Level */
#define FS454_VBI_BL_LVL	0x7C	/*! VBI Blank Level */
#define FS454_SOFT_RST		0x7E	/*! Encoder Soft Reset */
#define FS454_ENC_VER		0x7F	/*! Encoder Version */
#define FS454_WSS_CONFIG	0x80	/*! WSS Configuration Register */
#define FS454_WSS_CLK		0x81	/*! WSS Clock */
#define FS454_WSS_DATAF1	0x83	/*! WSS Data Field 1 */
#define FS454_WSS_DATAF0	0x86	/*! WSS Data Field 0 */
#define FS454_WSS_LNF1		0x89	/*! WSS Line Number Field 1 */
#define FS454_WSS_LNF0		0x8A	/*! WSS Line Number Field 0 */
#define FS454_WSS_LVL		0x8B	/*! WSS Level */
#define FS454_MISC_8D		0x8D	/*! Miscellaneous Bits Register 8D */
#define FS454_VID_CNTL0		0x92	/*! Video Control 0 */
#define FS454_HD_FP_SYNC	0x94	/*! Horiz. Front Porch & HSync Width */
#define FS454_HD_YOFF_BP	0x96	/*! HDTV Lum. Offset & Back Porch */
#define FS454_SYNC_DL		0x98	/*! Sync Delay Value */
#define FS454_LD_DET		0x9C	/*! DAC Load Detect */
#define FS454_DAC_CNTL		0x9E	/*! DAC Control */
#define FS454_PWR_MGNT		0xA0	/*! Power Management */
#define FS454_RED_MTX		0xA2	/*! RGB to YCrCb Matrix Red Coeff. */
#define FS454_GRN_MTX		0xA4	/*! RGB to YCrCb Matrix Green Coeff. */
#define FS454_BLU_MTX		0xA6	/*! RGB to YCrCb Matrix Blue Coeff. */
#define FS454_RED_SCL		0xA8	/*! RGB to YCrCb Scaling Red Coeff. */
#define FS454_GRN_SCL		0xAA	/*! RGB to YCrCb Scaling Green Coeff. */
#define FS454_BLU_SCL		0xAC	/*! RGB to YCrCb Scaling Blue Coeff. */
#define FS454_CC_FIELD_1	0xAE	/*! Closed Caption Field 1 Data */
#define FS454_CC_FIELD_2	0xB0	/*! Closed Caption Field 2 Data */
#define FS454_CC_CONTROL	0xB2	/*! Closed Caption Control */
#define FS454_CC_BLANK_VALUE	0xB4	/*! Closed Caption Blanking Value */
#define FS454_CC_BLANK_SAMPLE	0xB6	/*! Closed Caption Blanking Sample */
#define FS454_HACT_ST		0xB8	/*! HDTV Horizontal Active Start */
#define FS454_HACT_WD		0xBA	/*! HDTV Horizontal Active Width */
#define FS454_VACT_ST		0xBC	/*! HDTV Veritical Active Width */
#define FS454_VACT_HT		0xBE	/*! HDTV Veritical Active Height */
#define FS454_PR_PB_SCALING	0xC0	/*! Pr and Pb Relative Scaling */
#define FS454_LUMA_BANDWIDTH	0xC2	/*! Luminance Frequency Response */
#define FS454_QPR		0xC4	/*! Quick Program Register */

/*! Command register bits */

#define CR_GCC_CK_LVL		0x2000	/*! Graphics Controller switching lev */
#define CR_P656_LVL		0x1000	/*! Pixel Port Output switching level */
#define CR_P656_IN		0x0800	/*! Pixel Port In */
#define CR_P656_OUT		0x0400	/*! Pixel Port Out */
#define CR_CBAR_480P		0x0200	/*! 480P Color Bars */
#define CR_PAL_NTSCIN		0x0100	/*! PAL or NTSC input */
#define CR_SYNC_MS		0x0080	/*! Sync Master or Slave */
#define CR_FIFO_CLR		0x0040	/*! FIFO Clear */
#define CR_CACQ_CLR		0x0020	/*! CACQ Clear */
#define CR_CDEC_BP		0x0010	/*! Chroma Decimator Bypass */
#define CR_NCO_EN		0x0002	/*! Enable NCO Latch */
#define CR_SRESET		0x0001	/*! Soft Reset */

/*! Chip ID register bits */

#define FS454_CHIP_ID		0xFE05	/*! Chip ID register expected value */


static int fs454_read(__u32 reg, __u32 * word, __u32 len)
{
	int i;
	__u8 *wp = (__u8 *) word;

	*word = 0;

	int client = open(FS454_I2C_DEV, O_RDWR);
	ioctl(client, I2C_SLAVE_FORCE, FS454_I2C_ADDR);
	if (len == 1) *word = i2c_smbus_read_byte_data(client, reg);
    else if (len == 2) *word = i2c_smbus_read_word_data(client, reg);
	else if (len == 4) { *(__u16 *) word = i2c_smbus_read_word_data(client, reg);
	*((__u16 *) word + 1) = i2c_smbus_read_word_data(client, reg + 2);
} else
 return -EINVAL;
 close(client);

}

static int fs454_write(__u32 reg, __u32 word, __u32 len)
{
	int client = open(FS454_I2C_DEV, O_RDWR);
	ioctl(client, I2C_SLAVE_FORCE, FS454_I2C_ADDR);
	if (len == 1) return i2c_smbus_write_byte_data(client, reg, (__u8) word);
	else if (len == 2) return i2c_smbus_write_word_data(client, reg, (__u16) word);
	else if (len == 4) return i2c_smbus_write_block_data(client, reg, len,(__u8 *) & word);
	else return -EINVAL;
	close(client);
}

int memfd		= -1;
unsigned char *mmio	= NULL;
off_t pciaddr		= 0xf4000000;

#define i810_readb(mmio_base, where)                     \
        *((volatile char *) (mmio_base + where))

#define i810_readw(mmio_base, where)                     \
        *((volatile short *) (mmio_base + where))

#define i810_readl(mmio_base, where)                     \
        *((volatile int *) (mmio_base + where))

#define i810_writeb(mmio_base, where, val)                              \
        *((volatile char *) (mmio_base + where)) = (volatile char) val

#define i810_writew(mmio_base, where, val)                              \
        *((volatile short *) (mmio_base + where)) = (volatile short) val

#define i810_writel(mmio_base, where, val)                              \
        *((volatile int *) (mmio_base + where)) = (volatile int) val


void memClose(void) {
  if (mmio) munmap(mmio, 512 * 1024);
  if (memfd >= 0) close(memfd);
  memfd	= -1;
  mmio	= NULL;
}

void memOpen(void) {
  memfd	= open("/dev/mem", O_RDWR);
  if (memfd < 0) {
    int i	= errno;
    perror("/dev/mem");
    if (i == EACCES && geteuid() != 0)
      fprintf(stderr, "Must have access to `/dev/mem'.\n");
    memClose();
    exit(4);
  }

  mmio	= mmap(NULL, 512 * 1024, PROT_WRITE | PROT_READ, MAP_SHARED, memfd, pciaddr);
  if (!mmio) {
    memClose();
    exit(5);
  }
}

void i815_setup(void) {
	memOpen();

	i810_writel(mmio, 0x60000, 0x035f02cf); //HTOTAL
	i810_writel(mmio, 0x60004, 0x035f02cf); //HBLANK
	i810_writel(mmio, 0x60008, 0x033702f7); //HSYNC
	i810_writel(mmio, 0x6000c, 0x0270023f); //VTOTAL
	i810_writel(mmio, 0x60010, 0x0270023f); //VBLANK
	i810_writel(mmio, 0x60014, 0x02580256); //VSYNC
	i810_writel(mmio, 0x60018, 0xa0004047); //LCDTV_C
	i810_writel(mmio, 0x6001c, 0x02af033f); //OVRACT
	i810_writel(mmio, 0x60020, 0x00000000); //BCLRPAT

	memClose();
}

int fs454_setup(void)
{
	__u32 data;
	int i;
	
	// Focus Reset
	while (fs454_write(FS454_SOFT_RST, 0x01, 1));	
	while (fs454_write(FS454_SOFT_RST, 0x00, 1));	

	// Soft Reset Bit setzen
	while (fs454_write(FS454_CR, 0x2111, 2));

	for (i=0;i<0xc4;i++)
	{
		while (fs454_write(i,fs454_defaults[i], 1));
	}

	//PLL nderungen bernehmen
	while (fs454_write(FS454_CR, 0x2112, 2));
	while (fs454_write(FS454_CR, 0x2110, 2));

	//Soft Reset Bit lschen
	while (fs454_write(FS454_CR, 0x2110, 2));

	return 0;
}

void stv6412a_setup(void) {
	int client = open(STV6412_I2C_DEV, O_RDWR);
	ioctl(client, I2C_SLAVE_FORCE, 0x4a);
	i2c_smbus_write_byte_data(client, 0x03, 0xb6);
	close(client);
}

int main(int argc, char *argv[]) {
 i815_setup();
 fs454_setup();
 stv6412a_setup();
}

